#include <stdio.h>
#include <malloc.h>
#include "defines.h"

/* tree traversal */
#define LEFT(node) (node)->Left_Leaf
#define RIGHT(node) (node)->Right_Leaf
#define PREV(node) (node)->Prev_Leaf

/* traversal */
#define INSIDE(node) (node)->Left_Leaf
#define OUTSIDE(node) (node)->Prev_Leaf
#define ADJ_RIGHT(node) (node)->Right_Leaf
#define ADJ_LEFT(node) (node)->Prev_Leaf

#define TAG(node) (node)->tag_type

/* note: make sure root node does not move */

inline Tree *recurse_to_last_adj(Tree *tree)
{	
  if (tree != NULL)
    while (ADJ_RIGHT(tree) != NULL)
    {
	tree = ADJ_RIGHT(tree);
    }
    return tree;
}

int encountered_pl;

/* obfuscated code... */
inline int pull_out_nested_tags(Tree *location_a, Tree **location_b)
{	
    if (*location_b != NULL)
      if (INSIDE(*location_b) != NULL)
        if ((location_a->start_position == INSIDE(*location_b)->start_position) && 
	   (location_a->end_position == INSIDE(*location_b)->end_position))
	{
	   int i;
	   if ((TAG(INSIDE(*location_b)) == PL_TAG) || (TAG(location_a) == PL_TAG))	
	   	{ encountered_pl = 1; return 1; }
	   		
	    //printf("(%d, %d)\n",TAG(location_a), TAG(*location_b));
	    
	    if (encountered_pl)
	    {
		/* make sure only colors and fonts get pushed out of a pl */
		int t = TAG(*location_b);
		if ((t<r_TAG&&t>9)||(t>w_TAG))
			return 1;
	    }
	    
	    *location_b=INSIDE(*location_b);
	    i = TAG(*location_b);
	    TAG(*location_b) = TAG(location_a); 
	    TAG(location_a) = i;
	    //*location_b = INSIDE(*location_b);
	    return 1;
	}
      	
    return 0;
}

/* obfuscated code... */
void recurse(Tree *node)
{

    Tree *tmp, *nest_level_a, *nest_level_b;
    
    if (node != NULL)
    {
        tmp = PREV(node);
	nest_level_b = node;
	encountered_pl = 0; /* not sure if in right place */

	/* check if adjacent tags are the same */
	do
	{
	
	nest_level_a = tmp;
	
	do
	{

	  //printf ("(%d, %d)\n", TAG(node), TAG(tmp));

	    if (TAG(node) == TAG(tmp))
	    {
		Tree *new_pos, *no_more = node;

		/* eliminate adjacent node */
		tmp->end_position = node->end_position;
                ADJ_RIGHT(tmp) = ADJ_RIGHT(node);
		if (ADJ_RIGHT(node)!=NULL)
		    PREV(ADJ_RIGHT(node)) = tmp;
		node = ADJ_RIGHT(node);
		new_pos = recurse_to_last_adj(INSIDE(tmp));
		if (new_pos != NULL)
		  ADJ_RIGHT(new_pos) = INSIDE(no_more);
		else
		  new_pos = INSIDE(tmp); /* ??? */
		if (INSIDE(no_more) != NULL)
		    PREV(INSIDE(no_more)) = new_pos;
		free(no_more);
		nest_level_b = node;
		//printf("(%d,%d)\n", TAG(node), TAG(tmp));
	    }
	} while (pull_out_nested_tags(tmp, &nest_level_a));
	  
	} while (pull_out_nested_tags(node, &nest_level_b));
	
	if (node != NULL)
	{
            recurse(ADJ_RIGHT(node));
	    recurse(INSIDE(node));
	}
    }
}

void optimise_tree(Tree *tree)
{

    if (tree == NULL)
        return;
    if (ADJ_RIGHT(tree) != NULL)
        recurse(ADJ_RIGHT(tree));
    if (INSIDE(tree) != NULL)
        recurse(INSIDE(tree));	     
}

