/*
   Using flex for generic purposes
   Angelo
*/


%{
#include <stdio.h>
#include <stdlib.h>

#include "defines.h"

char *val_to_token[] = { "<0>", "<1>", "<2>", "<3>", "<4>", "<5>", "<6>", "<7>", "<8>",
			 "<9>", "<B>", "<EM>", "<I>", "<PL>", "<S>", "<TT>", "<U>", 
                         "<r>", "<g>","<b>", "<c>", "<m>", "<y>", "<k>", "<w>",
                         "</0>", "</1>", "</2>", "</3>", "</4>", "</5>", "</6>", "</7>", 
                         "</8>", "</9>", "</B>", "</EM>", "</I>", "</PL>", "</S>", 
                         "</TT>", "</U>", "</r>", "</g>","</b>", "</c>", "</m>", "</y>", 
                         "</k>", "</w>" };

int cur_tokens_on[END_TAG];

char plain_text[5000000];

Tree *root;
Tree *current;

int register_tag( int tag_type );
int insert_node( int tag_type );
void optimise_tree( Tree *current );
int print_tree( Tree *current );
int recreate_text( Tree *current );

/* booleans */
bool is_space = 0;
bool u_space = 0;

/* ints */
int u;
int size;
int color;


int em_track; /* last token -- any chars will invalidate this! */

int output_position = 0;
int current_file_position = 0;
%}

TEXT_CHAR [\x21-\x7e\x20\x0d\x0a\x09]

%%
"<B>"   insert_node( B_TAG );
"<EM>"  insert_node( EM_TAG );
"<I>"   insert_node( I_TAG );
"<PL>"  insert_node( PL_TAG );
"<S>"   insert_node( S_TAG );
"<TT>"  insert_node( TT_TAG );
"<U>"   insert_node( U_TAG );
\<[0-9]\> insert_node( atoi(&yytext[1]) );
"<r>"   insert_node( r_TAG ); 
"<g>"   insert_node( g_TAG );
"<b>"   insert_node( b_TAG );
"<c>"   insert_node( c_TAG );
"<m>"   insert_node( m_TAG );
"<y>"   insert_node( y_TAG );
"<k>"   insert_node( k_TAG );
"<w>"   insert_node( w_TAG );
"</B>"  insert_node( BDONE );
"</EM>"  insert_node( EMDONE );
"</I>"   insert_node( IDONE );   
"</PL>"  insert_node( PLDONE );
"</S>"   insert_node( SDONE );
"</TT>"  insert_node( TTDONE ); 
"</U>"   insert_node( UDONE );
\<"/"[0-9]\> insert_node( atoi(&yytext[2])+DIFF );
"</r>"   insert_node( rDONE );
"</g>"   insert_node( gDONE );
"</b>"   insert_node( bDONE );
"</c>"   insert_node( cDONE );
"</m>"   insert_node( mDONE );
"</y>"   insert_node( yDONE );
"</k>"   insert_node( kDONE );
"</w>"   insert_node( wDONE );
{TEXT_CHAR} {
           //   printf( "text char VALUE: %d\n", int(yytext[0]) ); 

                if ( yytext[0] != ' ' && yytext[0] != '\t' &&
                     yytext[0] != '\n' && yytext[0] != '\r' )
                {
                  plain_text[current_file_position] = yytext[0];
                  current_file_position++;    
                  is_space = 0;
                  em_track = 0;
                }
                else
                {
                 if ( is_space == 0 || cur_tokens_on[TT_TAG] != OFF ||
                      (cur_tokens_on[U_TAG] != OFF && u_space == OFF) )
                 {
                   plain_text[current_file_position] = yytext[0];
                   current_file_position++;
                 }
               
                 is_space = (cur_tokens_on[TT_TAG] == ON ) ? 0 : 1;

                 if (cur_tokens_on[U_TAG] == ON )
                   u_space = ON;
                 else
                   u_space = OFF;
                }                     
            }
%%

int main()
{
  int i;
 
  em_track = 0;

  root = (Tree *)malloc( sizeof(Tree) );
  current = root;
  root->tag_type = EMPTY;
  root->start_position = EMPTY;
  root->end_position = EMPTY;

  for ( i = 0; i < END_TAG; i++ )
    cur_tokens_on[i] = OFF;

  yylex();

//  printf("Root\n");
//  print_tree( root );

  plain_text[current_file_position+1] = '\0';
  //printf("\nrecreated = \n");
//  optimise_tree( root );
  recreate_text( root );

  while ( output_position < current_file_position )
  {
    printf( "%c", plain_text[output_position] );
    output_position++;
  }

  /* MAY HAVE TO BE REMOVED */
  printf( "\n");

  return 1;
}



// Used for removing needless tags
int register_tag( int tag_type )
{
  if ( tag_type < END_TAG )
  { 
    cur_tokens_on[tag_type]++;

    if ( tag_type != EM_TAG && tag_type != U_TAG &&
         tag_type != EMDONE && tag_type != UDONE )
      return cur_tokens_on[tag_type];
  }
  else
  {
    cur_tokens_on[tag_type-DIFF]--;   

    if ( tag_type != EM_TAG && tag_type != U_TAG &&
         tag_type != EMDONE && tag_type != UDONE )
      return cur_tokens_on[tag_type-DIFF]+1;
  }

  if ( tag_type == U_TAG || tag_type == UDONE )
  { 
    u_space = OFF;
    is_space = OFF;

    if ( tag_type == U_TAG )
    { 
      if ( cur_tokens_on[U_TAG] <= 3 )
        return 1;
      else
        return 0;
    }
    else
    {
      if ( cur_tokens_on[U_TAG]+1 <= 3 )
        return 1;
      else
        return 0;
    }
  }

  if ( tag_type == EM_TAG || tag_type == EMDONE )
  {
    if ( cur_tokens_on[PL_TAG] == 0 )
      em_track++;
    else 
      return 0;

    if ( em_track > 2 && tag_type == EM_TAG )
    {     
       if ( cur_tokens_on[EM_TAG] == 2 )
       {
         current->start_position = EMPTY;
         current->end_position = EMPTY;
         current->tag_type = EMPTY;

         cur_tokens_on[EM_TAG] = 0;
         return 0; 
       }
       
      return 1;
    }
    else
    if ( tag_type == EMDONE && cur_tokens_on[EM_TAG] != 0 )
    {
       if ( cur_tokens_on[EM_TAG] == -2 )
         cur_tokens_on[EM_TAG] = 0;

       return 0;
    }

    if ( cur_tokens_on[S_TAG] > 0 )
      return 0;
    else
      return 1;
  }
  return 1;
}


int insert_node( int tag_type )
{
   if ( register_tag(tag_type) == 1 )
   {
     if ( tag_type < END_TAG )
     {

      if ( current->end_position != EMPTY )
      {
        if ( current->tag_type != EMPTY )
        {
          current->Right_Leaf = (Tree *) malloc( sizeof(Tree) );
          current->Right_Leaf->Prev_Leaf = current;
          current = current->Right_Leaf;
        }
        current->tag_type = tag_type;
        current->start_position = current_file_position;
        current->end_position = EMPTY;
      }
      else
      {
        if ( current->tag_type != EMPTY )
        {
          current->Left_Leaf = (Tree *) malloc( sizeof(Tree) );
          current->Left_Leaf->Prev_Leaf = current;
          current = current->Left_Leaf;
        }
        current->tag_type = tag_type;
        current->start_position = current_file_position;
        current->end_position = EMPTY;
      }
    }
    else
    {   
      Tree *temp = current;

      while ( temp->tag_type != (tag_type - DIFF) || temp->end_position != EMPTY )
      { 
        if ( temp->Prev_Leaf == NULL ) 
          exit(-1);

     //   printf (" temp_tag: %d\n tag: %d\n", temp->tag_type, tag_type-DIFF );

        temp = temp->Prev_Leaf;
      }
 
      temp->end_position = current_file_position;
      current = temp;
    }
    return 1;
  }
  return -1;
}

int print_tree( Tree *current )
{
  if ( current->tag_type != EMPTY )
    printf( "Tag: %d\n   start_pos: %d\n   end_pos: %d\n", current->tag_type,
             current->start_position, current->end_position );

  if ( current->Left_Leaf != NULL )
  { 
    printf( "Left Leaf\n" );
    print_tree( current->Left_Leaf );
  }
  if ( current->Right_Leaf != NULL )
  {
    printf( "Right Leaf\n" );
    print_tree ( current->Right_Leaf );
  }

  return 1;
}


int recreate_text( Tree *current )
{
   while( output_position < current->start_position )
  {
    printf( "%c", plain_text[output_position] );
    output_position++;
  }

  if ( current->start_position == output_position )
    printf("%s", val_to_token[current->tag_type] );
  
  if ( current->Left_Leaf != NULL )
    recreate_text( current->Left_Leaf );
 
  while ( output_position < current->end_position )
  {
    printf( "%c", plain_text[output_position] );
    output_position++;
  }

  if ( current->end_position == output_position )
    printf( "%s", val_to_token[current->tag_type+set_diff_end] );

  if ( current->Right_Leaf != NULL )
    recreate_text( current->Right_Leaf );

  return 1;
}

